# -*- coding: utf-8 -*-
from collections import Counter
from sklearn.feature_extraction.text import CountVectorizer
from collections.abc import Iterable
from types import GeneratorType


__all__ = ["NGramFrequenzy"]


class NGramFrequenzy(object):
    """ Ngram frequencies from corpus.
    Args:
        corpus (iterable): Corpus as a sequence of docs each of type string
        frequency (dict or Counter, optional): ngram frequencies

    Raises:
        ValueError: Submitted frequency has to be either dict or Counter object

    """
    def __init__(self, corpus=[], frequency={}, **kwargs):
        if corpus:
            self.__ngram_freq = self.__extract_ngram_frequency(corpus=corpus, **kwargs)
        elif frequency:
            if isinstance(frequency, dict) or isinstance(frequency, GeneratorType):
                self.__ngram_freq = Counter(frequency)
            elif isinstance(frequency, Counter):
                self.__ngram_freq = frequency
            else:
                raise ValueError("Submitted frequency has to be either dict or Counter object")
        else:
            self.__ngram_freq =  Counter()

    @staticmethod
    def __extract_ngram_frequency(corpus, **kwargs):
        """ Count occurences for each word in the whole corpus.
            Arg:
                corpus (iterable): Series of documents each formatted as string
                
            Returns:
                pd.Series: Counts indexed by its respective word
        """
        cv = CountVectorizer(**kwargs)
        # get bag-of-words as sparse matrix
        bow = cv.fit_transform( corpus )
        # obtain list of feature names
        vocab = list( cv.get_feature_names() )
        # # word occurence for each column and collapse to vector
        word_counts = bow.sum(axis=0).A1
        # map word to its counts
        freq_distribution = Counter( dict( zip( vocab, word_counts) ) )
        return freq_distribution

    @classmethod
    def from_corpus(cls, corpus, **kwargs):
        """ Generate ngram frequencies from corpus

            Args:
                corpus (iterable): Series of documents each formatted as string

            Returns:
                NGramFrequenzy
        """
        if isinstance(corpus, list):
            return cls( frequency=cls.__extract_ngram_frequency(corpus, **kwargs) )
        else:
            raise  ValueError("Corpus has to be a list object")

    @classmethod
    def from_frequency(cls, frequency):
        """ Instantiate class by ngram frequencies

            Args:
                frequency (dict or Counter, optional): ngram frequencies

            Returns:
                NGramFrequenzy
        """
        if isinstance(frequency, dict) or isinstance(frequency, Counter):
            return cls( frequency = frequency )
        else:
            raise ValueError("N-Gram Frequency has to be either dict or Counter object")

    @property
    def degree(self):
        if self.is_empty():
            return 0
        else:
            for ngram in self.keys():
                return ngram.count(" ") + 1

    @property
    def total_frequency(self):
        if self.is_empty():
            return 0
        else:
            return sum( self.__ngram_freq.values() )

    @property
    def total_ngrams(self):
        if self.is_empty():
            return 0
        else:
            return len( self.__ngram_freq.keys() )

    def __repr__(self):
        if self.__ngram_freq:
            return str(self.__ngram_freq)

    def __str__(self):
        if self.__ngram_freq:
            return str( dict(self.__ngram_freq) )
        else:
            return str( dict() )

    def __add__(self, other):
        return NGramFrequenzy(frequency=self.__ngram_freq + other.__ngram_freq)

    def __iter__(self):
        for ngram, count in self.__ngram_freq.items():
            yield ngram, count

    def __getitem__(self, val):
        if isinstance(val, str):
            return self.__ngram_freq[val]
        elif isinstance(val, list):
            return self.search_ngrams(query=val)
        else:
            raise ValueError("Subcription value has to be either str or list")

    def keys(self):
        """ Iterator over ngrams """
        return self.__ngram_freq.keys()

    def values(self):
        """ Iterator over counts """
        return self.__ngram_freq.values()        

    def items(self):
        """ Iterator over ngram, counts tuples """
        return self.__ngram_freq.items()

    def search_ngrams(self, query, normalize=False, smoothing=1):
        """ Search ngrams which match the query from start.

            Args:
                query (str): query as string with blank as delimiter between token
                normalize (bool, optional): Normalize ngram frequencies. Defaults to False.

            Returns:
                NGramFrequenzy
        """
        parsed_query = query.split(" ") if isinstance(query, str) else query
        word_num = len(parsed_query)
        query_result = NGramFrequenzy(frequency={
            ngram: count
            for ngram, count in self.__ngram_freq.items() 
            if ngram.split(" ")[ :word_num ] == parsed_query
        })
        if not normalize:
            return query_result
        else:
            # if 1-gram take total frequency and do not apply smoothing
            if query_result.degree == 1:
                total = self.total_frequency
                smoothing = 1
            else: 
                total = query_result.total_frequency
            return NGramFrequenzy( frequency= { ngram: (count / total) * smoothing for ngram, count in query_result.items() } )

    def most_common(self, top_n=1, counts=True):
        """ Obtain most common ngrams and their respective counts in descending order.

        Args:
            top_n (int, optional): Number of top most ngrams. Defaults to 1.
            counts (bool, optional): Whether to also include counts. Defaults to True.

        Returns:
            list: If counts, then list of ngram with their respective counts as tuples,
                  else only ngrams
        """
        top_ngram_scores = self.__ngram_freq.most_common(top_n)
        if counts:
            return top_ngram_scores
        else:
            return list( ngram_score[0] for ngram_score in top_ngram_scores )            
            
    def _endswith(self, ending):
        return NGramFrequenzy(frequency={
            ngram: count
            for ngram, count in self.__ngram_freq.items() 
            if ngram.split(" ")[ -1 ] == ending
        })

    def is_empty(self):
        """ Check if there are any ngram frequency counts.

        Returns:
            bool
        """
        if not self.__ngram_freq.keys():
            return True
        else:
            return False

# TODO:
# - try to inherit from Counter
# - make compatible with generator
# - change object repr
# - implement collection class FrequencyCollection consting of multiple NGramFrequency 

if __name__ == "__main__":
    test_corpus = [
        "let us see were this project leads us",
        "we are having great fun so far",
        "we are having even greater fun now",
        "we are actively developing",
        "it is getting tougher but it is still fun",
        "this project teaches us how to construct test cases",
        "this project teaches us how to construct a package",
        "this project teaches us how to construct a setup tool",
        "this project teaches us how to build an api documentations",
    ]   
    ngfreq = NGramFrequenzy(corpus=test_corpus, ngram_range=(2,2))

    print(ngfreq.is_empty())
    print(ngfreq.degree)
    print(ngfreq.total_frequency)

    # print(ngfreq["are having"])
    # print( ngfreq.most_common(2, counts=True) )
    print(ngfreq.search_ngrams("are", normalize=True, smoothing=0.4))



import numpy as np
from collections import Counter

from .ngram import NGramFrequenzy

__all__ = ["NaiveNGram", "StupidBackoff"]


class BaseStatisticalModel(object):
    """Base class for all statistical models.
    """

    def __init__(self, n_max=2, frequencies=None):
        self.n_max = n_max
        self.model_frequencies = frequencies
    
    def _generate_freq(self, corpus):
        self.model_frequencies = { 
            n: NGramFrequenzy(corpus=corpus, ngram_range=( n, n ) ) 
            for n in range(1, self.n_max + 1)
        }

    @classmethod
    def __get_classname(cls):
        return cls.__name__

    def __make_ngram_stats(self):
        default = f"{ self.__get_classname() }: n_max = { self.n_max }"
        if self.model_frequencies:
            ngram_stats = "\n".join([
                f"  {n}-grams: count={ len( freq.keys() ) }, freq={ sum( freq.values() ) }" 
                 for n, freq in self.model_frequencies.items()
            ])
            return f"{default}\n{ngram_stats} "    
        else:
            return f"{default}, not fitted"

    def __repr__(self):
        return self.__make_ngram_stats()

    def __add__(self, other):
        left_max_n = len(self.model_frequencies.keys())
        right_max_n = len(other.model_frequencies.keys())
        common_max_n = max( left_max_n, right_max_n )
        return BaseStatisticalModel(frequencies={
            n: self.model_frequencies.get(n, NGramFrequenzy() ) + other.model_frequencies.get(n, NGramFrequenzy() )
            for n in range(1, common_max_n + 1)
        })

    def summary(self):
        """ Make summary statistics of fitted model

            Returns:
                str: summary as formatted string
        """
        return self.__make_ngram_stats()

    def ngram_frequency(self, ngram_seq, normalize=False):
        """ Obtain frequency of N-Gram from fitted model.

            Args:
                ngram_seq (str): N-Gram as a string
            
            Returns:
                int: Frequency if N-Gram
        """
        if self.model_frequencies:
            ngram_seq_parsed = ngram_seq.split(" ")
            ngram_seq_len = len(ngram_seq_parsed)

            if ngram_seq_len in self.model_frequencies:
                # search for ngrams with matching degree
                return self.model_frequencies[ ngram_seq_len ].search_ngrams(query=ngram_seq_parsed, normalize=normalize)
            else:
                return 0
        else:
            return 0


class NaiveNGram(BaseStatisticalModel):
    """ Model which determines next word of a sequence by finding N-Gram with most matching words. 
        It does not apply any smoothing to probability or scores (equivalent to Stupid Backoff with alpha=1).

        Args:
            n_max (int): Maximum n-gram degree
            threshold (int): Lowest n-gram degree to search in. If threshold reached, default to empty NGramFrequency.
    """
    def __init__(self, n_max, threshold=0, **kwargs):
        super(NaiveNGram, self).__init__(n_max, **kwargs)
        assert threshold >= 0
        self.threshold = threshold

    def _recursive_search(self, parsed_query, n, **kwargs):
        search_result = self.model_frequencies[ n ].search_ngrams(query=parsed_query, **kwargs)
        if not search_result.is_empty():
            return search_result
        else:
            n -= 1
            if n > self.threshold:
                # enter recursion with first word removed from query
                return self._recursive_search( parsed_query[ 1: ], n, **kwargs)
            else:
                return NGramFrequenzy()

    def _query_model(self, query, **kwargs):
        """ Query model to obtain longest matching ngram.

            Args:
                query (str): 

            Returns:
                NGramFrequenzy
        """
        parsed_query = query.split(" ")
        query_len = len(parsed_query)
        if query_len > self.n_max - 1:
            # if query longer than n_max + 1, then adjust 
            adjusted_query = parsed_query[ -(self.n_max -1): ]
            return self._recursive_search(adjusted_query, n=len(adjusted_query) + 1, **kwargs)
        else:
            return self._recursive_search(parsed_query, n=len(parsed_query) + 1, **kwargs)

    def fit(self, corpus):
        """ Fit model by generating word frequencies from input corpus.

            Args:
                corpus (iterable): Corpus as a sequence of docs each of type string
            
            Returns: 
                self
        """
        self._generate_freq(corpus)
        return self

    def score(self, queries, completions):
        """ Score completion for a given query.

            Args:
                queries (iterable): sequence of queries as strings
                completions (iterable): sequence of completions as strings for a given query

            Returns:
                list: score for each completion for a query
        """
        scores = []
        for query, comp in zip(queries, completions):
            search_result = self._query_model(query, normalize=True)._endswith( comp )
            if not search_result.is_empty():
                pred = search_result.most_common(1, counts=True)[0][1]
                scores.append( pred )
            else:
                scores.append(0)
        return scores

    def predict(self, queries):
        """ Model predictions based on input queries

            Args:
                queries (iterable): sequence queries each of type string

            Returns:
                list: prediction for each query as list of words each of type string
        """
        predictions = []
        for query in queries:
            search_result = self._query_model(query)
            if search_result.is_empty():
                predictions.append( np.NaN)
            else:
                # take last word ngram
                pred = search_result.most_common(1, counts=False)[0].split(" ")[-1]
                predictions.append( pred )
        return predictions

    def predict_proba(self, queries, top_n=1):
        """ Model predictions and its conditional probability based on input queries.

            Args:
                queries (iterable): sequence queries each of type string

            Returns:
                list: prediction for each query as list of words each of type string
        """
        probas = []
        for query in queries:
            search_result = self._query_model(query, normalize=True)
            if search_result.is_empty():
                probas.append( np.NaN )
            else:
                pred = search_result.most_common(1, counts=True)[0]
                probas.append( pred )
        return probas


class StupidBackoff(NaiveNGram):
    """ Stupid Backoff model as described in its original paper (https://www.aclweb.org/anthology/D07-1090.pdf)
        It recursively assigns a score to a word depending on its context which is further smoothed by parameter alpha.

        Args:
            n_max (int): Maximum n-gram degree
            threshold (int): Lowest n-gram degree to search in. If threshold reached, default to empty NGramFrequency.
            alpha (float): Smoothing parameter within range (0,1]
    """    
    def __init__(self, n_max, alpha=0.4, threshold=0, **kwargs):
        super(StupidBackoff, self).__init__(n_max, threshold, **kwargs)
        assert 0 < alpha <= 1
        self.alpha = alpha

    def _recursive_search(self, parsed_query, n, threshold=0, **kwargs):
        search_result = self.model_frequencies[ n ].search_ngrams(
            query=parsed_query, 
            smoothing=self.alpha**(self.n_max - n),
            **kwargs
        )
        if not search_result.is_empty():
            return search_result
        else:
            n -= 1
            if n > self.threshold:
                # enter recursion with first word removed from query
                return self._recursive_search( parsed_query[ 1: ], n, **kwargs)
            else:
                return NGramFrequenzy()            



# TODO: 
#     - additional language models (markov chain, backoff etc)
#     - methods for exploring word freq
#     - improve efficiency of ngram-search (e.g make ngram comparison more efficient by comparing lists instead of strings)
#     - add type checking

if __name__ == "__main__":
    test_corpus = [
        "let us see were this project leads us",
        "we are having great fun so far",
        "we are actively developing",
        "it is getting tougher but it is still fun",
        "this project teaches us how to construct test cases",
    ]        
    sb = StupidBackoff(n_max=3, alpha=0.4, threshold=0)
    sb.fit( test_corpus )
    infer_doc = [
        # "let us see were that project",
        # "they are",
        # "it is",
        "it should be",
        # "we",
    ]
    print(sb)
    print(sb.predict(infer_doc))
    print(sb.predict_proba(infer_doc))
    print(sb.score(["they are"], ["actively"]))



# -*- coding: utf-8 -*-


from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name="statLM",
    version="0.0.2",
    author="Raphael Redmer",
    license="MIT",
    author_email="ra.redmer@outlook.com",
    description="Language models to predict words",
    long_description=readme,
    long_description_content_type="text/x-rst",
    url="https://github.com/RaRedmer/statLM",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=required,
)



# -*- coding: utf-8 -*-

import unittest
import numpy as np


from statLM import statistical_models as sm

class BasicTestSuite(unittest.TestCase):
    """Basic test cases."""

    def __init__(self, *args, **kwargs):
        super(BasicTestSuite, self).__init__(*args, **kwargs)
        self.test_corpus = [
            "let us see were this project leads us",
            "we are having great fun so far",
            "we are actively developing",
            "it is getting tougher but it is still fun",
            "this project teaches us how to construct test cases",
        ]
        self.test_queries = [
            "let us see were that project",
            "we are",
            "it is",
            "it should be",
            "we",
        ]
        self.test_completions = ["leads", "actively", "getting", "us", "are"]

    def test_naive_ngram(self):
        # Naive Ngram
        nn = sm.NaiveNGram(n_max=3, threshold=1)
        nn.fit( self.test_corpus )
        self.assertEqual(
            nn.predict(self.test_queries), 
            ['leads', 'actively', 'getting', np.NaN, "are"]
        )
        self.assertEqual(
            nn.predict_proba(self.test_queries), 
            [('project leads', 0.5), ('we are actively', 0.5),
             ('it is getting', 0.5), np.NaN, ('we are', 1.0)]
        )
        self.assertEqual(
            nn.score( 
                queries=self.test_queries,
                completions=self.test_completions
            ),
            [0.5, 0.5, 0.5, 0, 1.0]
        )

    def test_naive_ngram_no_threshold(self):
        # Naive Ngram
        nn = sm.NaiveNGram(n_max=3, threshold=0)
        nn.fit( self.test_corpus )
        self.assertEqual(
            nn.predict(self.test_queries), 
            ['leads', 'actively', 'getting', "us", "are"]
        )
        self.assertEqual(
            nn.predict_proba(self.test_queries), 
            [('project leads', 0.5), ('we are actively', 0.5),
             ('it is getting', 0.5), ("us", 3/37), ('we are', 1.0)]
        )        

    def test_stupid_backoff(self):
        sb = sm.StupidBackoff(n_max=3, alpha=0.4, threshold=1)
        sb.fit( self.test_corpus )
        self.assertEqual(
            sb.predict(self.test_queries), 
            ['leads', 'actively', 'getting', np.NaN, "are"]
        )
        self.assertEqual(
            sb.predict_proba(self.test_queries), 
            [('project leads', 0.2), ('we are actively', 0.5), ('it is getting', 0.5), np.NaN, ('we are', 0.4)]
        )        
        self.assertEqual(
            sb.score( 
                queries=self.test_queries,
                completions=self.test_completions
            ),
            [0.2, 0.5, 0.5, 0, 0.4]
        )        

    def test_stupid_backoff_no_threshold(self):
        sb = sm.StupidBackoff(n_max=3, alpha=0.4, threshold=0)
        sb.fit( self.test_corpus )
        self.assertEqual(
            sb.predict(self.test_queries), 
            ['leads', 'actively', 'getting', "us", "are"]
        )
        self.assertEqual(
            sb.predict_proba(self.test_queries), 
            [('project leads', 0.2), ('we are actively', 0.5), ('it is getting', 0.5), ("us", 3/37), ('we are', 0.4)]
        )        
        self.assertEqual(
            sb.score( 
                queries=self.test_queries,
                completions=self.test_completions
            ),
            [0.2, 0.5, 0.5, 3/37, 0.4]
        )   

# TODO: 
#     - construct more test cases
    # - automate test via github action

if __name__ == '__main__':
    unittest.main()